#!/usr/bin/perl

#   Copyright (C) 2005 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

# Please email any bugs, comments, and/or additions to this file to:
# ishiura-gcc@ksc.kwansei.ac.jp

# This file was written by Youki Uchiyama. (ishiura-gcc@ksc.kwansei.ac.jp)


use strict;
use warnings;
use Getopt::Std;

our $limits_path;

package main;
my %configVariables;

my $options = &readOptions();
&generateFileAndDirectory();
&checkArgument();
my @templateFilenames = @ARGV;
my $templateFileProcessor = TemplateFileProcessor::new('TemplateFileProcessor',
									\@templateFilenames, $configVariables{TESTDIR}, $options);
TemplateFileProcessor::generateTestsuite($templateFileProcessor);

sub checkArgument{
	if (scalar(@ARGV) == 0) {
		print "arguments not exist";
		&dispHelp();
	}
	foreach (@ARGV) {
		if ($_ !~ /.tt$/) {
			print "$_\n";
			print "illegal arguments\n";
			print @ARGV;
			&dispHelp();
		}
	}
}
sub readOptions{
	my %opt;
	getopts('c:hfi',\%opt) or die &dispHelp();
	if ($opt{h}) {
		&dispHelp();
	}
	if($opt{f} and $opt{i}){
		print "illegal options\n";
		print "not supported same use i & f options\n";
		exit;
	}
	if(defined $opt{c}){
		;
	}elsif(-e "./tgen.cnf"){
		$opt{c} = "tgen.cnf";
	}elsif(-e "/etc/testgen/tgen.cnf"){
		$opt{c} = "/etc/testgen/tgen.cnf";
	}else{
		print "not found config file!\n";
		exit;
	}
	&readConfigFile($opt{c});
	return \%opt;
}
sub readConfigFile{
	my $configFilename = shift;
	print "config file : $configFilename\n";
	open(CONFIG_FILE, $configFilename) or die "failed to open $configFilename\n";
	while (<CONFIG_FILE>) {
		chomp;
		if($_ =~ /^\s*$/){
			next;
		}
		my @elements = split /\:/, $_;
		my ($name, $value);
		if (scalar @elements == 2) {
			$name = shift @elements;
			$value = shift @elements;
		} elsif (scalar @elements == 1 or $elements[1] =~ /\s+/) {
			$name = shift @elements;
			$value = "";
		} else {
			print "error in config file\n";
			exit;
		}
		checkReserved($name);
		$configVariables{$name} = $value;
	}
	close CONFIG_FILE;
}
sub checkReserved{
	my $name = shift;
	my $reserved = 0;
	my @reservationWords = qw(CC TOOL TARGET OBJ LIBS LOG LDFLAGS
TESTDIR SIMULATOR_NAME FILSET INT_BIT CHAR_BIT SHORT_BIT LONG_BIT);
	foreach (@reservationWords) {
		if ($name eq $_) {
			$reserved = 1;
		}
	}
	if($reserved){
		return $reserved;
	}else{
		print "syntax error\n";
		print "name : $name\n";
		exit;
	}
}
sub generateFileAndDirectory{
	&generateTestDirectory();
	my @tmp = @ARGV;
	my $tmp = shift @tmp;
	my @tmp2 = split /\//, $tmp;
	pop @tmp2;
	my $path = join '/', @tmp2;
	if($path =~ /^\s*$/){
		$path = ".";
	}
	system "cp $path/gcc_template.exp ./$configVariables{TESTDIR}/config/$configVariables{TARGET}\.exp";
	&generateRuntest();
	&generateLimitsHeader();
}
#¸ߤʤХƥȥǥ쥯ȥ
sub generateTestDirectory{
	&checkDirectoryName();
	if (! -d "./$configVariables{TESTDIR}") {
		mkdir "./$configVariables{TESTDIR}", 0755
		  or die "cannot make directory './$configVariables{TESTDIR}'";
	}
	if(-e "/usr/share/testgen/lib.tar.gz"){
		system "tar zxf /usr/share/testgen/lib.tar.gz -C ./$configVariables{TESTDIR}/";
	}else{
		system "tar zxf ./lib.tar.gz -C ./$configVariables{TESTDIR}/";
	}
	if(! -e "./$configVariables{TESTDIR}/config"){
		mkdir "./$configVariables{TESTDIR}/config", 0755
		  or die "cannnot make directory './$configVariables{TESTDIR}/config'";
	}
	if(! -e "./$configVariables{TESTDIR}/$configVariables{LOG}"){
		mkdir "./$configVariables{TESTDIR}/$configVariables{LOG}", 0755
		  or die "cannot make directory './$configVariables{TESTDIR}/$configVariables{LOG}'";
	}
}
#ǥ쥯ȥ̾Ʊ̾Υե뤬Х顼ɽ
sub checkDirectoryName{
	if (-f "./$configVariables{TESTDIR}") {
		print "ʥե뤬¸ߤޤ>>./$configVariables{TESTDIR}\n";
		exit;
	}
}
sub generateRuntest{
	open(RUNTESTFILE, ">./$configVariables{TESTDIR}/RUNTEST")
	  or die "failed to open ./$configVariables{TESTDIR}/RUNTEST";
	print RUNTESTFILE << "RUNTEST_END";
#!/bin/sh
set -x

TOOL=$configVariables{TOOL}
CC=$configVariables{CC}
TARGET=$configVariables{TARGET}
OBJDIR=$configVariables{OBJ}
LOGDIR=$configVariables{LOG}
LIBS=$configVariables{LIBS}
LDFLAGS=$configVariables{LDFLAGS}
SIMULATOR_NAME=$configVariables{SIMULATOR_NAME}
FILESET=$configVariables{FILSET}

while [ \$1 ]; do
	if [ -di = \$1 ]; then
	DI_DIR=\${2:+"-di \${2\%\%/}"}
	shift
	elif [ -h = \$1 ]; then
	  echo "Usage: \$0 [-di dirname] [expfiles]" 1>\&2
	  exit 1
	else
	EXPFILES=\${1:+\${EXPFILES} \$1}
	fi
	shift
done

runtest --tool \$TOOL --target \$TARGET --objdir \${OBJDIR}/\${1+\"\$\@\"\} --outdir \$LOGDIR CC=\$CC LDFLAGS=\$LDFLAGS LIBS=\$LIBS FILESET=\$FILESET SIMULATOR_NAME=\$SIMULATOR_NAME  \$DI_DIR \$EXPFILES
RUNTEST_END
	system "chmod +x ./$configVariables{TESTDIR}/RUNTEST";
}
sub generateLimitsHeader{
	open(LIMITS_HEADER, ">./$configVariables{TESTDIR}/lib/limits.h")
	  or die "failed to open ./$configVariables{TESTDIR}/lib/limits.h";
	$limits_path = "./$configVariables{TESTDIR}/lib/limits.h";
	my $char_max = 2 ** ($configVariables{CHAR_BIT} - 1) - 1; 
	my $char_min = (2 ** ($configVariables{CHAR_BIT} - 1)) * (-1);
	my $uchar_max = 2 ** $configVariables{CHAR_BIT} - 1;
	my $short_max = 2 ** ($configVariables{SHORT_BIT} - 1) - 1;
	my $short_min = (2 ** ($configVariables{SHORT_BIT} - 1)) * (-1);
	my $ushort_max = 2 ** $configVariables{SHORT_BIT} - 1;
	my $int_max = 2 ** ($configVariables{INT_BIT} - 1) - 1;
	my $int_min = (2 ** ($configVariables{INT_BIT} - 1)) * (-1);
	my $uint_max = 2 ** $configVariables{INT_BIT} - 1;
	my $long_max = 2 ** ($configVariables{LONG_BIT} - 1) - 1;
	my $long_min = (2 ** ($configVariables{LONG_BIT} - 1)) * (-1);
	my $ulong_max = 2 ** $configVariables{LONG_BIT} - 1;
	print LIMITS_HEADER << "LIMITS_END";
#define __CHAR_MAX $char_max
#define __CHAR_MIN $char_min
#define __UCHAR_MAX $uchar_max
#define __SHORT_MAX $short_max
#define __SHORT_MIN $short_min
#define __USHORT_MAX $ushort_max
#define __INT_MAX $int_max
#define __INT_MIN $int_min
#define __UINT_MAX $uint_max
#define __LONG_MAX $long_max
#define __LONG_MIN $long_min
#define __ULONG_MAX $ulong_max
LIMITS_END
}
sub dispHelp{
	print << "HELP_END";
usage : tgen.pl [-h] [-c CONFIG_FILE] [-f|-i] templateFile ...
-c CONFIG_FILE : specification config file
-h             : display this help message
-f             : generate test only float type
-i             : generate test only integer type
HELP_END
	exit;
}

package TemplateFileProcessor;
sub new{
	my $pkg = shift;
	my $templateFilenames = shift; #reference
	my $testDir = shift;
	my $options = shift;
	bless {templateFilenames=>$templateFilenames, testDir=>$testDir, options=>$options}, $pkg;
}
sub generateTestsuite{
	my $this = shift;
	foreach my $templateFilename(@{$this->{templateFilenames}}) {
		(my $dirname = $templateFilename) =~ s/\.tt$//;
		my @tmp = split /\//, $dirname;
		$dirname = pop @tmp;
		my $directory = "./$this->{testDir}/$dirname";
		my $templateFile = TemplateFile::new('TemplateFile', $templateFilename, $directory,
											 $this->{options});
		$templateFile->readTemplate();
		if(! $this->{options}->{n}){
			TemplateFile::generateTestDirectory($templateFile);
		}
	}
}

package TemplateFile;
sub new{
	my $pkg = shift;
	my $templateFilename = shift;
	my $directory = shift;
	my $options = shift; #reference
	my %globalDefine;
	my @blocks;
	my @joins;
	my @lines;
	my @expFilenames;
	my $index = 0;
	my $fileNumber = -1;

	open(TEMPLATE, "$templateFilename") or die "failed to open $templateFilename";
	print "reading : $templateFilename\n";
	while (<TEMPLATE>) {
		push(@lines, $_);
	}
	close TEMPLATE;

	bless {templateFilename=>$templateFilename, directory=>$directory, blocks=>\@blocks,
		 globalDefine=>\%globalDefine, joins=>\@joins, lines=>\@lines, index=>$index,
		 expFilenames=>\@expFilenames, fileNumber=>$fileNumber, options=>$options}, $pkg;
}
sub readTemplate{
	my $this = shift;
	while (scalar @{$this->{lines}} > $this->{index}) {
		if ($this->currentLine() =~ /^\@file\s+\S+/) {
			$this->addFileBlock();
		} elsif ($this->currentLine() =~ /^\@for\s+\S+\s+/) {
			$this->addForBlock();
		} elsif ($this->currentLine() =~ /^\@define\s+(\w+)\s+\'(.*)\'/) {
			$this->addDefine();
		} elsif ($this->currentLine() =~ /^\@multiline\s+(\w+)/) {
			$this->addMultilineBlock();
		} elsif ($this->currentLine() =~ /^\@join\s+/) {
			$this->addJoin();
		} elsif ($this->currentLine() =~ /^\@run\s+/) {
			$this->addExpfiles();
		} elsif ($this->currentLine() =~ /^\s*$/) {
			#Ԥ̵
			$this->nextLine();
		} elsif ($this->currentLine() =~ /^\@comment/) {
			while ($this->currentLine() !~ /^\@end_comment/) {
				$this->nextLine();
			}
			$this->nextLine();
		} else {
			print "in TemplateFile::readTemplate\n";
			$this->dispNumberCurrentLine();
		}
	}
}
sub addFileBlock{
	my $this = shift;
	my @block;
	while ($this->currentLine() !~ /^\@end_file/) {
		if($this->currentLine() =~ /^\@/ and $this->currentLine !~ /^\@file/){
			print "in TemplateFile::addFileBlock\n";
			$this->dispNumberCurrentLine();
		}
		push(@block, $this->currentLine());
		$this->nextLine();
	}
	push(@block, $this->currentLine()); #add @end_file
	my $fileBlock = FileBlock::new('FileBlock', \@block);
	$fileBlock->readFileBlock();
	push(@{$this->{blocks}}, $fileBlock);
	$this->nextLine();
}
sub addForBlock{
	my $this = shift;
	my @block;
	my $nest = 0;
	push(@block, $this->currentLine());
	$this->nextLine();
	while($this->currentLine() =~ /^\s*$/){
		$this->nextLine();
	}
	while ($nest != 0 or $this->currentLine() !~ /^\@end_for/) {
		if ($this->currentLine() =~ /^\@end_for/) {
			$nest--;
		} elsif ($this->currentLine() =~ /^\@for/) {
			$nest++;
		} elsif ($this->currentLine() =~ /^\@/ and $this->currentLine() !~ /^\@file/
				 and $this->currentLine() !~ /^\@end_file/) {
			print "in TepmlateFile::addForBlock\n";
			$this->dispNumberCurrentLine();
		}
		push(@block, $this->currentLine());
		$this->nextLine();
	}
	push(@block, $this->currentLine()); #add @end_for
	my $forBlock = ForBlock::new('ForBlock', \@block);
	$forBlock->readForBlock();
	push(@{$this->{blocks}},$forBlock);
	$this->nextLine();
}
sub addDefine{
	my $this = shift;
	$this->{globalDefine}->{$1} = $2;
	$this->nextLine();
}
sub addMultilineBlock{
	my $this = shift;
	my @block;
	my $name;
	$name = $1;
	$this->nextLine();
	while ($this->currentLine() !~ /^\@end_multiline/) {
		push(@block, $this->currentLine());
		$this->nextLine();
	}
	$this->{globalDefine}->{$name} = join("", @block);
	$this->nextLine();
}
sub addJoin{
	my $this = shift;
	chomp(my $line = $this->currentLine());
	$line =~ s/\@join\s+//;
	push(@{$this->{joins}}, $line);
	$this->nextLine();
}
sub addExpfiles{
	my $this = shift;
	chomp(my $line = $this->currentLine());
	$line =~ s/^\@run\s+//;
	foreach (split /\s+/, $line) {
		push @{$this->{expFilenames}}, $_;
	}
	$this->nextLine();
}
sub dispError{
	my $this = shift;
	print "unsupported @... found\n";
	print "$this->currentLine()\n";
	exit;
}
sub currentLine{
	my $this = shift;
	$this->{lines}->[$this->{index}];
}
sub nextLine{
	my $this = shift;
	if($this->{index} > scalar @{$this->{lines}}){
		print 'б@endĤޤ\n';
		print "ե뽪üãޤ\n";
		exit;
	}
	$this->{index}++;
}
sub dispNumberCurrentLine{
	my $this = shift;
	$this->nextLine();
	print "line: $this->{index}\n";
	print "error: $this->{lines}->[$this->{index}-1]\n";
	exit;
}
sub generateTestDirectory{
	my $this = shift;
	#create directory ... g.1-1 etc
	if(! -e "$this->{directory}"){
		mkdir "$this->{directory}" , 0755 or die "cannot make directory $this->{directory}";
	}else{
		system "rm -rf $this->{directory}/*";
	}
	print "generating : $this->{directory}\n";
	foreach(@{$this->{blocks}}){
		if(! defined $_){
			print "error: in TemplateFile::generateTestDirectory\n";
			print "exist undefined value in @{$this->{blocks}}\n";
			exit;
		}
	}
	my $numberOfBlock;
	my %groups;
	foreach my $refBlock(@{$this->{blocks}}) {
		my @references = $refBlock->getCode();
		my @testFilenames = $refBlock->getFilenames();
		my $numberOfBlock = 0;
		foreach my $block(@references){
			my $autoFilename = $this->createTestFilename($testFilenames[$numberOfBlock]);
			$numberOfBlock++;
			if ($refBlock->getGroupName ne "") {
				$groups{$refBlock->getGroupName()} .= "$autoFilename:";
			}
			if(defined $this->{options}->{i} and $autoFilename =~ /^f/){
				next;
			}elsif(defined $this->{options}->{f} and $autoFilename !~ /^f/){
				next;
			}
			#create an test file
			$this->makeTestFile($autoFilename, $block);
		}
	}
	#add lines to FILSET from @joins
	if (scalar @{$this->{joins}} != 0) {
		$this->generateFilset(\%groups);
	}
	#generate exp file
	my $expNumber = 1;
  my $tmp = $this->{templateFilename};
  my @tmp = split /\//, $tmp;
  pop @tmp;
  my $path = join '/', @tmp;
  if($path =~ /^\s*$/){
    $path = ".";
  }
	foreach(@{$this->{expFilenames}}){
		system "cp $path/$_ $this->{directory}/$expNumber-$_";
		$expNumber++;
	}
  #cp testout.h
  system "cp $path/testout.h $this->{directory}/";
	#cp limits.h
	system "cp $limits_path $this->{directory}/";
}
sub createTestFilename{
	my $this = shift;
	my $filename = shift;
	$this->{fileNumber}++;
	if ($filename =~ /([^\?]*)(\?*)(.*)/) {
		my $questionNumber;
		if( ($questionNumber = length $2) != 0 ){
			my $tmp1 = $1;
			my $tmp3 = $3;
			my $fileNumber = sprintf "%${questionNumber}s", $this->{fileNumber};
			$fileNumber =~ s/\s/0/g;
			return "$tmp1$fileNumber$tmp3";
		} else {
			return $filename;
		}
	}
}
sub makeTestFile{
	my $this = shift;
	my $autoFilename = shift;
	my $block = shift; #reference
	if (-e "$this->{directory}/$autoFilename") {
		print "error: exist same name file\n";
		exit;
	}
	open(FILE, ">$this->{directory}/$autoFilename")
or die "failed to open '$this->{directory}/$autoFilename'";
	foreach my $line (@$block) {
		if ($line =~ /\$(\w+)/ ) {
			my $element = $1;
			if (defined $this->{globalDefine}->{$element}) {
				$line =~ s/\$$element/$this->{globalDefine}->{$element}/g;
			}
		}
		print FILE $line;
	}
	close FILE;
}
sub generateFilset{
	my $this = shift;
	my $groups = shift; #reference
	my $filenamesOfSameGroup;
	my @filset;
	foreach my $line (@{$this->{joins}}) {
		my $beforeGroup = undef;
		my $group;
		foreach my $groupName (split / /, $line) {
			if (! defined $groups->{$groupName}) {
				print "not exist $groupName\n";
				exit;
			}
			my @filenames = split(/:/, $groups->{$groupName});
			my $currentGroup = Group::new('Group', $groupName, \@filenames);
			if (defined $beforeGroup) {
				$currentGroup->{refGroup} = $beforeGroup;
				$beforeGroup = $currentGroup;
			} else {
				$currentGroup->{line} = $line;
				$beforeGroup = $currentGroup;
			}
			$group = $currentGroup;
		}
		push @filset, $group->getFilsetLine();
	}
	open(FILSET, ">>$this->{directory}/FILSET") or die "failed to open 'FILSET'";
	foreach (@filset) {
		print FILSET "$_\n";
	}
	close FILSET;
}

package Group;
sub new{
	my $pkg = shift;
	my $groupName = shift;
	my $filenames = shift; #reference
	my $line;
	my $refGroup;
	bless {line=>$line, groupName=>$groupName, filenames=>$filenames, refGroup=>$refGroup}, $pkg;
}
sub getFilsetLine{
	my $this = shift;
	my @lines;
	if(defined $this->{line}){
		push @lines,$this->{line};
	}else{
		@lines = $this->{refGroup}->getFilsetLine();
	}
	my @transferedLines;
	my @filenames = @{$this->{filenames}};
	foreach my $line (@lines) {
		foreach my $filename (@filenames) {
			my $tmp = $line;
			$tmp =~ s/$this->{groupName}/$filename/;
			push @transferedLines, $tmp;
		}
	}
	return @transferedLines;
}

package ForBlock;
sub new{
	my $pkg = shift;
	my $variableName;
	my @values;
	my $referBlock;
	my $tmp = shift; #reference
	my @lines = @$tmp;
	my $index = 0;
	bless {variableName=>$variableName, values=>\@values, referBlock=>$referBlock,
		  lines=>\@lines, index=>$index}, $pkg;
}
sub readForBlock{
	my $this = shift;
	if ($this->currentLine() =~ /^\@for\s+(\w+)\s+/) {
		$this->{variableName} = $1;
		my $argument = $';
		my @tmp = split(/\'/, $argument);
		if(scalar @tmp <= 0){
			print "in ForBlock::readForBlock\n";
			$this->dispNumberCurrentLine();
		}
		for (my $i=1; $i<scalar @tmp; $i+=2) {
			push(@{$this->{values}}, $tmp[$i]);
		}
	}
	$this->nextLine();
	if ($this->currentLine() =~ /^\@for/) {
		$this->addForBlock();
	} elsif ($this->currentLine() =~ /^\@file/) {
		$this->addFileBlock();
	}
}
sub addForBlock{
	my $this = shift;
	my @block;
	my $nest = 0;
	push(@block, $this->currentLine());
	$this->nextLine();
	while($this->currentLine() =~ /^\s*$/){
		$this->nextLine();
	}
	while (! ($nest == 0 and $this->currentLine() =~ /^\@end_for/)) {
		if ($this->currentLine() =~ /^\@end_for/) {
			$nest--;
		} elsif ($this->currentLine() =~ /^\@for/) {
			$nest++;
		} elsif ($this->currentLine() =~ /^\@/ and $this->currentLine() !~ /^\@file/
				 and $this->currentLine() !~ /^\@end_file/) {
		}
		push(@block, $this->currentLine());
		$this->nextLine();
	}
	push(@block, $this->currentLine()); #add @end_for
	my $forBlock = new ForBlock(\@block);
	$forBlock->readForBlock();
	$this->{referBlock} = $forBlock;
	$this->nextLine();
}
sub addFileBlock{
	my $this = shift;
	my @block;
	while ($this->currentLine() !~ /^\@end_file/) {
		push(@block, $this->currentLine());
		$this->nextLine();
	}
	push(@block, $this->currentLine()); #add @end_file
	my $fileBlock = new FileBlock(\@block);
	$fileBlock->readFileBlock();
	$this->{referBlock} = $fileBlock;
	$this->nextLine();
}
sub nextLine{
	my $this = shift;
	$this->{index}++;
}
sub currentLine{
	my $this = shift;
	$this->{lines}->[$this->{index}];
}
sub dispNumberCurrentLine{
	my $this = shift;
	$this->nextLine();
	print "line: $this->{index}\n";
	print "error: $this->{lines}->[$this->{index}-1]\n";
	exit;
}
sub getCode{
	my $this = shift;
	if(! defined $this->{referBlock}){
		print "error: in ForBlock::getCode\n";
		exit;
	}
	my @references = $this->{referBlock}->getCode();
	my @transferedReferences;
	foreach my $block(@references) {
		foreach my $value (@{$this->{values}}) {
			my @transferedBlock;
			foreach my $line (@$block) {
				(my $transferedLine = $line) =~ s/\$$this->{variableName}/$value/g;
				push(@transferedBlock, $transferedLine);
			}
			push(@transferedReferences, \@transferedBlock);
		}
	}
	return @transferedReferences;
}
sub getFilenames{
	my $this = shift;
	if(! defined $this->{referBlock}){
		print "error: in ForBlock::getFilenames\n";
		exit;
	}
	my @testFilenames = $this->{referBlock}->getFilenames();
	my @transferedFilenames;
	foreach my $testFilename(@testFilenames) {
		foreach my $value (@{$this->{values}}) {
			(my $transferedFilename = $testFilename) =~ s/\$$this->{variableName}/$value/g;
			push(@transferedFilenames, $transferedFilename);
		}
	}
	return @transferedFilenames;
}
sub getGroupName{
	my $this = shift;
	$this->{referBlock}->getGroupName();
}

package FileBlock;
sub new{
	my $pkg = shift;
	my $lines = shift; #reference
	my @testFilenames;
	my $groupName = "";
	my $index = 0;

	bless{lines=>$lines, testFilenames=>\@testFilenames, groupName=>$groupName,
		  index=>$index}, $pkg;
}
sub readFileBlock{
	my $this = shift;
	if ($this->currentLine() =~ /^\@file\s+([\?\w\$]+\.\w+)\s+(\w+)/ ) {
		push(@{$this->{testFilenames}}, $1);
		$this->{groupName} = $2;
	} elsif ($this->currentLine() =~ /^\@file\s+(\S+)/ ) {
		push(@{$this->{testFilenames}}, $1);
	}else{
		print 'error: illegal @file format\n';
		exit;
	}
	shift @{$this->{lines}};
	pop @{$this->{lines}};
}
sub getCode{
	my $this = shift;
	my @references;
	push @references, [@{$this->{lines}}];
	return @references;
}
sub getFilenames{
	my $this = shift;
	return @{$this->{testFilenames}};
}
sub getGroupName{
	my $this = shift;
	return $this->{groupName};
}
sub currentLine{
	my $this = shift;
	$this->{lines}->[$this->{index}];
}
sub nextLine{
	my $this = shift;
	$this->{index}++;
}
